# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Client components for RFC9176_

Currently, this only contains the :mod:`register` module, which helps utilizing
the registration interface of a resource directory.

.. _RFC9176: https://datatracker.ietf.org/doc/html/rfc9176
"""
